@@ -17,6 +17,8 @@ from tqdm import tqdm
 from einops import rearrange
 from einops_exts import check_shape, rearrange_many
 
+from video_diffusion_pytorch.text import tokenize, bert_embed, BERT_MODEL_DIM
+
 # helpers functions
 
 def exists(x):
@@ -48,6 +50,53 @@ def prob_mask_like(shape, prob, device):
     else:
         return torch.zeros(shape, device = device).float().uniform_(0, 1) < prob
 
+def is_list_str(x):
+    if not isinstance(x, (list, tuple)):
+        return False
+    return all([type(el) == str for el in x])
+
+# relative positional bias
+
+class RelativePositionBias(nn.Module):
+    def __init__(
+        self,
+        heads = 8,
+        num_buckets = 32,
+        max_distance = 128
+    ):
+        super().__init__()
+        self.num_buckets = num_buckets
+        self.max_distance = max_distance
+        self.relative_attention_bias = nn.Embedding(num_buckets, heads)
+
+    @staticmethod
+    def _relative_position_bucket(relative_position, num_buckets = 32, max_distance = 128):
+        ret = 0
+        n = -relative_position
+
+        num_buckets //= 2
+        ret += (n < 0).long() * num_buckets
+        n = torch.abs(n)
+
+        max_exact = num_buckets // 2
+        is_small = n < max_exact
+
+        val_if_large = max_exact + (
+            torch.log(n.float() / max_exact) / math.log(max_distance / max_exact) * (num_buckets - max_exact)
+        ).long()
+        val_if_large = torch.min(val_if_large, torch.full_like(val_if_large, num_buckets - 1))
+
+        ret += torch.where(is_small, n, val_if_large)
+        return ret
+
+    def forward(self, n, device):
+        q_pos = torch.arange(n, dtype = torch.long, device = device)
+        k_pos = torch.arange(n, dtype = torch.long, device = device)
+        rel_pos = rearrange(k_pos, 'j -> 1 j') - rearrange(q_pos, 'i -> i 1')
+        rp_bucket = self._relative_position_bucket(rel_pos, num_buckets = self.num_buckets, max_distance = self.max_distance)
+        values = self.relative_attention_bias(rp_bucket)
+        return rearrange(values, 'i j h -> h i j')
+
 # small helper modules
 
 class EMA():
@@ -110,9 +159,9 @@ class PreNorm(nn.Module):
         self.fn = fn
         self.norm = LayerNorm(dim)
 
-    def forward(self, x):
+    def forward(self, x, **kwargs):
         x = self.norm(x)
-        return self.fn(x)
+        return self.fn(x, **kwargs)
 
 # building block modules
 
@@ -205,17 +254,35 @@ class Attention(nn.Module):
         self.to_qkv = nn.Linear(dim, hidden_dim * 3, bias = False)
         self.to_out = nn.Linear(hidden_dim, dim, bias = False)
 
-    def forward(self, x):
+    def forward(
+        self,
+        x,
+        pos_bias = None,
+        attend_self_only = False
+    ):
         qkv = self.to_qkv(x).chunk(3, dim = -1)
-        q, k, v = rearrange_many(qkv, '... n (h d) -> ... h n d', h = self.heads)
-        q = q * self.scale
 
-        sim = einsum('... h i d, ... h j d -> ... h i j', q, k)
-        sim = sim - sim.amax(dim = -1, keepdim = True).detach()
-        attn = sim.softmax(dim = -1)
+        if not attend_self_only:
+            q, k, v = rearrange_many(qkv, '... n (h d) -> ... h n d', h = self.heads)
+            q = q * self.scale
+
+            sim = einsum('... h i d, ... h j d -> ... h i j', q, k)
+
+            if exists(pos_bias):
+                sim = sim + pos_bias
+
+            sim = sim - sim.amax(dim = -1, keepdim = True).detach()
+            attn = sim.softmax(dim = -1)
+
+            out = einsum('... h i j, ... h j d -> ... h i d', attn, v)
+            out = rearrange(out, '... h n d -> ... n (h d)')
+        else:
+            # it was not clear in the paper
+            # but i assume by arresting the attention across time
+            # would mean simply having each token attend to itself
+            # which would be equivalent to passing that token's values through to the output
+            out = qkv[-1]
 
-        out = einsum('... h i j, ... h j d -> ... h i d', attn, v)
-        out = rearrange(out, '... h n d -> ... n (h d)')
         return self.to_out(out)
 
 # model
@@ -227,7 +294,9 @@ class Unet3D(nn.Module):
         cond_dim = None,
         out_dim = None,
         dim_mults=(1, 2, 4, 8),
-        channels = 3
+        channels = 3,
+        attn_heads = 8,
+        use_bert_text_cond = False
     ):
         super().__init__()
         self.channels = channels
@@ -243,7 +312,9 @@ class Unet3D(nn.Module):
             nn.Linear(dim * 4, dim)
         )
 
-        self.has_cond = exists(cond_dim)
+        self.has_cond = exists(cond_dim) or use_bert_text_cond
+        cond_dim = BERT_MODEL_DIM if use_bert_text_cond else cond_dim
+
         self.null_cond_emb = nn.Parameter(torch.randn(1, cond_dim)) if self.has_cond else None
 
         cond_dim = time_dim + int(cond_dim)
@@ -253,7 +324,14 @@ class Unet3D(nn.Module):
 
         num_resolutions = len(in_out)
         conv_next = partial(ConvNextBlock, time_emb_dim = cond_dim)
-        temporal_attn = lambda dim: EinopsToAndFrom('b c f h w', 'b (h w) f c', Attention(dim))
+
+        # temporal attention and its relative positional encoding
+
+        temporal_attn = lambda dim: EinopsToAndFrom('b c f h w', 'b (h w) f c', Attention(dim, heads = attn_heads))
+
+        self.time_rel_pos_bias = RelativePositionBias(heads = attn_heads, max_distance = 32) # realistically will not be able to generate that many frames of video... yet
+
+        # modules for all layers
 
         for ind, (dim_in, dim_out) in enumerate(in_out):
             is_last = ind >= (num_resolutions - 1)
@@ -261,7 +339,7 @@ class Unet3D(nn.Module):
             self.downs.append(nn.ModuleList([
                 conv_next(dim_in, dim_out, norm = ind != 0),
                 conv_next(dim_out, dim_out),
-                Residual(PreNorm(dim_out, SpatialLinearAttention(dim_out))),
+                Residual(PreNorm(dim_out, SpatialLinearAttention(dim_out, heads = attn_heads))),
                 Residual(PreNorm(dim_out, temporal_attn(dim_out))),
                 Downsample(dim_out) if not is_last else nn.Identity()
             ]))
@@ -269,7 +347,7 @@ class Unet3D(nn.Module):
         mid_dim = dims[-1]
         self.mid_block1 = conv_next(mid_dim, mid_dim)
 
-        spatial_attn = EinopsToAndFrom('b c f h w', 'b f (h w) c', Attention(mid_dim))
+        spatial_attn = EinopsToAndFrom('b c f h w', 'b f (h w) c', Attention(mid_dim, heads = attn_heads))
 
         self.mid_spatial_attn = Residual(PreNorm(mid_dim, spatial_attn))
         self.mid_temporal_attn = Residual(PreNorm(mid_dim, temporal_attn(mid_dim)))
@@ -282,7 +360,7 @@ class Unet3D(nn.Module):
             self.ups.append(nn.ModuleList([
                 conv_next(dim_out * 2, dim_in),
                 conv_next(dim_in, dim_in),
-                Residual(PreNorm(dim_in, SpatialLinearAttention(dim_in))),
+                Residual(PreNorm(dim_in, SpatialLinearAttention(dim_in, heads = attn_heads))),
                 Residual(PreNorm(dim_in, temporal_attn(dim_in))),
                 Upsample(dim_in) if not is_last else nn.Identity()
             ]))
@@ -293,7 +371,12 @@ class Unet3D(nn.Module):
             nn.Conv3d(dim, out_dim, 1)
         )
 
-    def forward_with_cond_scale(self, *args, cond_scale = 2., **kwargs):
+    def forward_with_cond_scale(
+        self,
+        *args,
+        cond_scale = 2.,
+        **kwargs
+    ):
         logits = self.forward(*args, null_cond_prob = 0., **kwargs)
         if cond_scale == 1:
             return logits
@@ -301,7 +384,14 @@ class Unet3D(nn.Module):
         null_logits = self.forward(*args, null_cond_prob = 1., **kwargs)
         return null_logits + (logits - null_logits) * cond_scale
 
-    def forward(self, x, time, cond = None, null_cond_prob = 0.):
+    def forward(
+        self,
+        x,
+        time,
+        cond = None,
+        null_cond_prob = 0.,
+        focus_on_the_present = True  # sounds quite spiritual
+    ):
         assert not (self.has_cond and not exists(cond)), 'cond must be passed in if cond_dim specified'
 
         t = self.time_mlp(time) if exists(self.time_mlp) else None
@@ -316,17 +406,19 @@ class Unet3D(nn.Module):
 
         h = []
 
+        time_rel_pos_bias = self.time_rel_pos_bias(x.shape[2], device = x.device)
+
         for convnext, convnext2, spatial_attn, temporal_attn, downsample in self.downs:
             x = convnext(x, t)
             x = convnext2(x, t)
             x = spatial_attn(x)
-            x = temporal_attn(x)
+            x = temporal_attn(x, pos_bias = time_rel_pos_bias, attend_self_only = focus_on_the_present)
             h.append(x)
             x = downsample(x)
 
         x = self.mid_block1(x, t)
         x = self.mid_spatial_attn(x)
-        x = self.mid_temporal_attn(x)
+        x = self.mid_temporal_attn(x, pos_bias = time_rel_pos_bias, attend_self_only = focus_on_the_present)
         x = self.mid_block2(x, t)
 
         for convnext, convnext2, spatial_attn, temporal_attn, upsample in self.ups:
@@ -334,7 +426,7 @@ class Unet3D(nn.Module):
             x = convnext(x, t)
             x = convnext2(x, t)
             x = spatial_attn(x)
-            x = temporal_attn(x)
+            x = temporal_attn(x, pos_bias = time_rel_pos_bias, attend_self_only = focus_on_the_present)
             x = upsample(x)
 
         return self.final_conv(x)
@@ -468,6 +560,11 @@ class GaussianDiffusion(nn.Module):
 
     @torch.no_grad()
     def sample(self, cond = None, cond_scale = 1., batch_size = 16):
+        device = next(self.denoise_fn.parameters()).device
+
+        if is_list_str(cond):
+            cond = bert_embed(tokenize(cond)).to(device)
+
         batch_size = cond.shape[0] if exists(cond) else batch_size
         image_size = self.image_size
         channels = self.channels
@@ -498,12 +595,16 @@ class GaussianDiffusion(nn.Module):
             extract(self.sqrt_one_minus_alphas_cumprod, t, x_start.shape) * noise
         )
 
-    def p_losses(self, x_start, t, cond = None, noise = None):
+    def p_losses(self, x_start, t, cond = None, noise = None, **kwargs):
         b, c, f, h, w = x_start.shape
         noise = default(noise, lambda: torch.randn_like(x_start))
 
         x_noisy = self.q_sample(x_start=x_start, t=t, noise=noise)
-        x_recon = self.denoise_fn(x_noisy, t, cond = cond)
+
+        if is_list_str(cond):
+            cond = bert_embed(tokenize(cond)).to(x_start.device)
+
+        x_recon = self.denoise_fn(x_noisy, t, cond = cond, **kwargs)
 
         if self.loss_type == 'l1':
             loss = F.l1_loss(noise, x_recon)
